""" Propagators for classical particle dynamics """

import mpmath as mp
import numpy as np
import sympy as sp
from liouville import SPLPQP

class Propagator:
    """ Base class for all propagators for classical particle dynamics """
    name = 'generic'
    def __init__(self, syst, tstart, tend, tstep, substep=None, pbc=False,
                 tqdm=False):
        self.q = syst.q0
        self.p = syst.p0
        self.q0 = syst.q0
        self.p0 = syst.p0
        self.syst = syst
        self.tstart = tstart
        self.tend = tend
        self.tstep = tstep
        self.substep = substep
        self.en0 = self.syst.energy(self.q0, self.p0)
        assert self.tstart < self.tend
        nsteps = int(np.floor((self.tend-self.tstart)/self.tstep))+1
        # self.time = np.linspace(self.tstart, self.tend, nsteps)
        self.time = [tstart+(tend-tstart)/(nsteps-1)*s for s in range(nsteps)]
        self.sq = [self.q]
        self.sp = [self.p]
        self.pbc = pbc
        self.step = self.step_pbc if self.pbc else self.step_nopbc
        self.tqdm = tqdm

        if self.substep is not None:
            assert self.substep < self.tstep
            assert np.isclose(self.tstep % self.substep, 0)
            nsubsteps = int(np.floor(self.tstep/self.substep))
            self.subtime = np.linspace(self.substep, self.tstep, nsubsteps)

    def propagate(self):
        """ perform propagation of coordinate and momentum in time """
        from tqdm import tqdm
        (q, p) = (self.q, self.p)
        if self.substep is not None:
            its = tqdm(self.time[:-1]) if self.tqdm else self.time[:-1]
            for time in its:
                (qt, pt) = self.step(time, q, p)
                self.sq.extend(qt)
                self.sp.extend(pt)
                q = qt[-1]
                p = pt[-1]
            self.time = [t+st for t in self.time[:-1] for st in self.subtime]
            self.time.insert(0, 0)
        else:
            its = tqdm(self.time[1:], miniters=10) if self.tqdm else self.time[1:]
            for time in its:
                (q, p) = self.step(time, q, p)
                self.sq.append(q)
                self.sp.append(p)
    '''
    def step(self, time, coordinate, momentum):
        """ perform a time step - stub method """
        return (coordinate, momentum)
    '''
    def get_trajectory(self, step=1, dtype=np.float):
        """ return a trajectory - a time series of coordinate and momentum """
        time = np.array(self.time, dtype=dtype)[::step]
        coordinate = np.array(self.sq, dtype=dtype)[::step]
        momentum = np.array(self.sp, dtype=dtype)[::step]
        return time, coordinate, momentum

    def get_trajectory_3d(self, step=1):
        """ return a trajectory of coordinates grouped in 3-tuples """
        ti, qt, pt = self.get_trajectory(step)
        qt.shape = (len(qt), -1, 3)
        pt.shape = (len(pt), -1, 3)
        return ti, qt, pt

    def analyse(self, step=1, p='err'):
        """ calculate relative energy error or relative energy difference
            (p = 1, 2, 3, ..., inf), see refs. https://doi.org/10.2307/2683905
            https://en.wikipedia.org/wiki/Relative_change_and_difference """
        from itertools import islice
        if not hasattr(self, 'en'):
            qptup = islice(zip(self.sq, self.sp), 0, None, step)
            self.en = np.array([self.syst.energy(*t) for t in qptup])
        if p == 'inf':
            func = lambda x, y: (x-y)/np.maximum(abs(x), abs(y))
        elif p == 'err':
            func = lambda x, y: (x-y)/y
        elif isinstance(p, int) and p > 0:
            func = lambda x, y: (x-y)/(abs(x)**p+abs(y)**p)**(1/p)
        else:
            raise ValueError('invalid value for p: '+repr(p))
        self.er = np.array(func(self.en, self.en0), dtype=float)


class PolynomialPropagator(Propagator):
    """ Base class for polynomial classical propagators """

    name = 'polynomial'
    coeffs = None
    iln = None

    def __init__(self, liouville=SPLPQP, liou_params={}, restart=False,
                 restart_file='propag.pkl', fpprec='fixed', prec=None,
                 nterms=None, delta=1., **kwargs):
        super().__init__(**kwargs)
        self.liouville = liouville
        self.liou_params = liou_params
        self.restart = restart
        self.restart_file = restart_file
        self.fpprec = fpprec
        self.prec = prec
        assert nterms is not None and nterms > 0
        self.nterms = nterms
        assert delta > 0.
        self.delta = delta
        self.alpha = self.delta*self.tstep/2.
        self.efficiency = self.alpha/self.nterms

    def set_polynomial(self, **kwargs):
        """ initialize the polynomial (iln) object """
        if self.restart:
            self.iln = self.load(self.restart_file)
        else:
            params = {'fpprec': self.fpprec, **kwargs, **self.liou_params}
            self.iln = self.liouville(self.nterms, **params)

    def step_nopbc(self, time, qval, pval):
        """ perform a time step with initial coordinate and momentum """
        qvec, pvec = self.iln.get_val_float(qval, pval)
        qt = np.dot(self.coeffs, qvec)
        pt = np.dot(self.coeffs, pvec)
        return (qt, pt)

    def step_pbc(self, *args, **kwargs):
        """ perform a time step with periodic boundary conditions """
        wrap = lambda x, y: (self.syst.wrap_q(x), y)
        return wrap(*self.step_nopbc(*args, **kwargs))

    def load(self, filename='propag.pkl'):
        """ deserialize the propagator object and extract the iln object """
        import cloudpickle as pkl
        with open(filename, 'rb') as fh:
            that = pkl.load(fh)
        params = ['__class__', 'name', 'nterms', 'delta', 'tstep', 'liouville',
                  'liou_params', 'fpprec', 'prec']
        for key in params:
            attr = getattr(that, key)
            errstr = key + ': ' + str(getattr(self, key)) + ' != ' + str(attr)
            assert getattr(self, key) == attr, errstr
        assert self.syst.hamiltonian == that.syst.hamiltonian
        return that.iln

    def dump(self, filename='propag.pkl'):
        """ serialize and save the propagator object """
        import cloudpickle as pkl
        with open(filename, 'wb') as fh:
            pkl.dump(self, fh)


class TwoStepPolynomialPropagator(PolynomialPropagator):
    """ Two-step time-symmetric polynomial propagators """

    name = 'two-step polynomial'
    coeffs = None
    iln = None

    def __init__(self, q_1=None, p_1=None, **kwargs):
        """ the starter propagator takes a backward time step """
        super().__init__(**kwargs)
        self.parity = (self.nterms-1)%2
        self.prefact = 2*self.parity-1
        if q_1 is None or p_1 is None:
            starter = self.starter(**kwargs)
            starter.coeffs = [c*(-1)**i for i, c in enumerate(starter.coeffs)]
            self.q_1, self.p_1 = starter.step(None, self.q0, self.p0)
        else:
            self.q_1, self.p_1 = q_1, p_1

    def step_nopbc(self, time, qval, pval):
        """ perform a time step with initial coordinate and momentum """
        qvec, pvec = self.iln.get_val_float(qval, pval)
        qt = np.dot(self.coeffs, qvec) + self.prefact*self.q_1
        pt = np.dot(self.coeffs, pvec) + self.prefact*self.p_1
        self.q_1 = qval
        self.p_1 = pval
        return (qt, pt)


class SymmetricPolynomialPropagator(PolynomialPropagator):
    """ Self-starting one-step time-symmetric polynomial propagators """

    name = 'symmetric polynomial propagator'
    numeric = {'tool': 'subs', 'modules': None}

    def __init__(self, numeric=None, **kwargs):
        """
        -------------------------------------------------------------------
        fpprec  dtype       tool        modules     linalg      method
        -------------------------------------------------------------------
        fixed   np.float    subs        None        sympy       step_sympy
        fixed   np.float    theano      None        numpy       step_numpy
        fixed   np.Float    lambdify    numpy       numpy       step_numpy
        multi   mp.Float    subs        None        sympy       step_sympy
        multi   mp.Float    lambdify    mpmath      mpmath      step_mpmath
        -------------------------------------------------------------------
        """

        from generic import validate_numeric

        if numeric is not None:
            self.numeric = numeric
        tool = self.numeric['tool']
        assert tool in ['subs', 'lambdify', 'theano']
        if tool == 'subs':
            self.step_nopbc = self.step_sympy
        if tool == 'lambdify':
            if self.numeric['modules'] == 'mpmath':
                self.step_nopbc = self.step_mpmath
            else:
                self.step_nopbc = self.step_numpy
        if tool == 'theano':
            self.step_nopbc = self.step_numpy

        super().__init__(**kwargs)

        validate_numeric(self.fpprec, self.numeric)

        if self.fpprec == 'fixed':
            self.prec = np.finfo(float).precision

    def step_sympy(self, time, qval, pval):
        """ this version works with sympy objects """
        zval = sp.Matrix([qval, pval])
        qvec, pvec = self.iln.get_val_float(qval, pval)
        qval0 = self.coeffs.dot(qvec)
        pval0 = self.coeffs.dot(pvec)
        zval0 = sp.Matrix([qval0, pval0])
        gcon = zval + zval0
        zval1 = zval0.copy()
        residue = 1.
        # Newton iterations
        numiter = 0
        while residue > 10**(-self.prec/2):
            if numiter > self.maxiter:
                raise RuntimeError('maximum number of iterations exceeded')
            jacval = self.jacobian.xreplace(dict(zip(self.z, zval1))).evalf()
            gvar = self.gvar.xreplace(dict(zip(self.z, zval1))).evalf()
            correction = jacval.inv()*(gvar-gcon)
            zval1 = zval1 - correction
            residue = correction.norm()
            numiter += 1
        return (zval1[0], zval1[1])

    def step_mpmath(self, time, qval, pval):
        """ this version works with lambdify and mpmath objects """

        zval = mp.matrix([qval, pval])
        qvec, pvec = self.iln.get_val_float(qval, pval)
        qval0 = self.coeffs.dot(qvec)
        pval0 = self.coeffs.dot(pvec)
        zval0 = mp.matrix([qval0, pval0])
        gcon = zval + zval0
        zval1 = zval0.copy()
        residue = 1.
        # Newton iterations
        numiter = 0
        while residue > 10**(-self.prec/2):
            if numiter > self.maxiter:
                raise RuntimeError('maximum number of iterations exceeded')
            jacval = self.jacobian(*zval1)
            gvar = self.gvar(*zval1)
            correction = jacval**(-1)*(gvar-gcon)
            zval1 = zval1 - correction
            residue = mp.mnorm(correction, p='f')
            numiter += 1
        return (zval1[0], zval1[1])

    def step_numpy(self, time, qval, pval):
        """ this version works with theano/lambdify and numpy objects """
        zval = np.array([qval, pval])
        qvec, pvec = self.iln.get_val_float(qval, pval)
        qval0 = np.dot(self.coeffs, qvec)
        pval0 = np.dot(self.coeffs, pvec)
        zval0 = np.array([qval0, pval0])
        gcon = zval + zval0
        zval1 = zval0.copy()
        residue = 1.
        # Newton iterations
        numiter = 0
        while residue > 10**(-self.prec/2):
            if numiter > self.maxiter:
                raise RuntimeError('maximum number of iterations exceeded')
            jacval = np.array(self.jacobian(*zval1))
            gvar = np.array(self.gvar(*zval1))
            if self.numeric['tool'] == 'theano':
               jacval.shape = self.jshape
            correction = np.matmul(np.linalg.inv(jacval), gvar-gcon)
            zval1 = zval1 - correction
            residue = np.linalg.norm(correction)
            numiter += 1
        return (zval1[0], zval1[1])
